import os
import sys
from datetime import datetime,timedelta
import logging
import pathlib
import tempfile
import subprocess
import shutil
from typing import Union
from time import time

import numpy as np
import scipy as sp
#from numba import jit, prange
import netCDF4 as nc
from netCDF4 import Dataset
from matplotlib.transforms import Bbox
import seawater as sw
import xarray as xr

from pyschism.mesh.base import Nodes, Elements
from pyschism.mesh.vgrid import Vgrid

logger = logging.getLogger(__name__)

def convert_longitude(ds, bbox):
#https://stackoverflow.com/questions/53345442/about-changing-longitude-array-from-0-360-to-180-to-180-with-python-xarray
#Light_B's solution didn't generate the correct result
#Michael's solution works, but it takes significantly longer to write nc file (~30 mins compared with 5 mins)
#TODO: figure out why it takes much longer with the second method
    #lon_attr = ds.coords['lon'].attrs
    if bbox.xmin < 0:
        logger.info(f'Convert HYCOM longitude from [0, 360) to [-180, 180):')
        #ds.coords['lon'] = (ds.coords['lon'] + 180) % 360 - 180
        ds['_lon_adjusted'] = xr.where(ds['lon'] > 180, ds['lon'] - 360, ds['lon'])
    elif bbox.xmin > 0:
        logger.info(f'Convert HYCOM longitude from [-180, 180) to [0, 360): ')
        #ds.coords['lon'] = (ds.coords['lon'] + 360) % 360 - 180
        ds['_lon_adjusted'] = xr.where(ds['lon'] < 0, ds['lon'] + 360, ds['lon'])

    t0 = time()
    ds = (
        ds.swap_dims({'lon': '_lon_adjusted'})
        .sel(**{'_lon_adjusted': sorted(ds._lon_adjusted)})
        .drop('lon')
    )
    ds = ds.rename({'_lon_adjusted': 'lon'})
    #ds = ds.sortby(ds.lon)
    #ds.coords['lon'].attrs = lon_attr
    logger.info(f'swap dims took {time()-t0} seconds!')

    #make sure it is clipped to the bbox
    ds = ds.sel(lon=slice(bbox.xmin - 0.5, bbox.xmax + 0.5))

    return ds

def get_database(date, Bbox=None):
    if date >= datetime(2018, 12, 4):
        database = f'GLBy0.08/expt_93.0'
    elif date >= datetime(2018, 1, 1) and date < datetime(2018, 12, 4):
        database = f'GLBv0.08/expt_93.0'
    elif date >= datetime(2017, 10, 1) and date < datetime(2018, 1, 1):
        database = f'GLBv0.08/expt_92.9'
    elif date >= datetime(2017, 6, 1) and date < datetime(2017, 10, 1):
        database = f'GLBv0.08/expt_57.7'
    elif date >= datetime(2017, 2, 1) and date < datetime(2017, 6, 1):
        database = f'GLBv0.08/expt_92.8'
    elif date >= datetime(2016, 5, 1) and date < datetime(2017, 2, 1):
        database = f'GLBv0.08/expt_57.2'
    elif date >= datetime(2016, 1, 1) and date < datetime(2016, 5, 1):
        database = f'GLBv0.08/expt_56.3'
    elif date >= datetime(1994, 1, 1) and date < datetime(2016, 1, 1):
        database = f'GLBv0.08/expt_53.X/data/{date.year}'
    else:
        raise ValueError(f'No data fro {date}!') 
    return database

def get_idxs(date, database, bbox, lonc=None, latc=None):

    if date >= datetime.utcnow():
        date2 = datetime.utcnow() - timedelta(days=1)
        baseurl = f'https://tds.hycom.org/thredds/dodsC/{database}/FMRC/runs/GLBy0.08_930_FMRC_RUN_{date2.strftime("%Y-%m-%dT12:00:00Z")}?depth[0:1:-1],lat[0:1:-1],lon[0:1:-1],time[0:1:-1]'
    else:
        baseurl=f'https://tds.hycom.org/thredds/dodsC/{database}?lat[0:1:-1],lon[0:1:-1],time[0:1:-1],depth[0:1:-1]'

    ds=Dataset(baseurl)
    time1=ds['time']
    times=nc.num2date(time1,units=time1.units,only_use_cftime_datetimes=False)
    
    lon=ds['lon'][:]
    lat=ds['lat'][:]
    dep=ds['depth'][:]
    
    #check if hycom's lon is the same range as schism's
    same = True
    if not (bbox.xmin >= lon.min() and bbox.xmax <= lon.max()):
        same = False
        if lon.min() >= 0:
            logger.info(f'Convert HYCOM longitude from [0, 360) to [-180, 180):')
            idxs = lon>=180
            lon[idxs] = lon[idxs]-360
        elif lon.min() <= 0:
            logger.info(f'Convert HYCOM longitude from [-180, 180) to [0, 360):')
            idxs = lon<=0
            lon[idxs] = lon[idxs]+360

    lat_idxs=np.where((lat>=bbox.ymin-0.5)&(lat<=bbox.ymax+0.5))[0]
    lon_idxs=np.where((lon>=bbox.xmin-0.5) & (lon<=bbox.xmax+0.5))[0]
    lon=lon[lon_idxs]
    lat=lat[lat_idxs]
    #logger.info(lon_idxs)
    #logger.info(lat_idxs)
    lon_idx1=lon_idxs[0].item()
    lon_idx2=lon_idxs[-1].item()
    #logger.info(f'lon_idx1 is {lon_idx1}, lon_idx2 is {lon_idx2}')
    lat_idx1=lat_idxs[0].item()
    lat_idx2=lat_idxs[-1].item()
    #logger.info(f'lat_idx1 is {lat_idx1}, lat_idx2 is {lat_idx2}')
    
    if lonc is None:
        lonc = lon.mean()
    #logger.info(f'lonc is {lonc}')
    if latc is None:
        latc = lat.mean()
    #logger.info(f'latc is {latc}')
    x2, y2=transform_ll_to_cpp(lon, lat, lonc, latc)

    idxs=np.where( date == times)[0]
    #check if time_idx is empty
    if len(idxs) == 0:
        #If there is missing data, use the data from the next days, the maximum searching days is 3. Otherwise, stop.
        for i in np.arange(0,3):
            date_before=(date + timedelta(days=int(i)+1)) #.astype(datetime)
            logger.info(f'Try replacing the missing data from {date_before}')
            idxs=np.where(date_before == times)[0]
            if len(idxs) == 0:
                continue
            else:
                break
    if len(idxs) ==0:
        logger.info(f'No date for date {date}')
        sys.exit()
    time_idx=idxs.item()  

    ds.close()

    return time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2, same

def transform_ll_to_cpp(lon, lat, lonc=-77.07, latc=24.0):
    longitude=lon/180*np.pi
    latitude=lat/180*np.pi
    radius=6378206.4
    loncc=lonc/180*np.pi
    latcc=latc/180*np.pi
    lon_new=[radius*(longitude[i]-loncc)*np.cos(latcc) for i in np.arange(len(longitude))]
    lat_new=[radius*latitude[i] for i in np.arange(len(latitude))]

    return np.array(lon_new), np.array(lat_new)

def interp_to_points_3d(dep, y2, x2, bxyz, val):
    idxs = np.where(abs(val) > 10000)
    val[idxs] = float('nan')

    if not np.all(x2[:-1] <= x2[1:]):
        logger.info('x2 is not in stricitly ascending order! Sorting x2 and val')
        idxs = np.argsort(x2)
        x2 = x2[idxs]
        val = val[:, :, idxs]

    val_fd = sp.interpolate.RegularGridInterpolator((dep,y2,x2),np.squeeze(val),'linear', bounds_error=False, fill_value = float('nan'))
    val_int = val_fd(bxyz)
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        val_int[idxs] = sp.interpolate.griddata(bxyz[~idxs,:], val_int[~idxs], bxyz[idxs,:],'nearest')
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        logger.info(f'There is still missing value for {val}')
        sys.exit()
    return val_int

def interp_to_points_2d(y2, x2, bxy, val):
    idxs = np.where(abs(val) > 10000)
    val[idxs] = float('nan')

    if not np.all(x2[:-1] <= x2[1:]):
        logger.info('x2 is not in stricitly ascending order! Sorting x2 and val')
        idxs = np.argsort(x2)
        x2 = x2[idxs]
        val = val[:, idxs]

    val_fd = sp.interpolate.RegularGridInterpolator((y2,x2),np.squeeze(val),'linear', bounds_error=False, fill_value = float('nan'))
    val_int = val_fd(bxy)
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        val_int[idxs] = sp.interpolate.griddata(bxy[~idxs,:], val_int[~idxs], bxy[idxs,:],'nearest')
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        logger.info(f'There is still missing value for {val}')
        sys.exit()
    return val_int

def ConvertTemp(salt, temp, dep):
    nz = temp.shape[0]
    ny = temp.shape[1]
    nx = temp.shape[2]
    pr = np.ones(temp.shape)
    pre = pr*dep[:,None, None]
    Pr = np.zeros(temp.shape)
    ptemp = sw.ptmp(salt, temp, pre, Pr)*1.00024
    return ptemp

class OpenBoundaryInventory:

    def __init__(self, hgrid, vgrid=None):
        self.hgrid = hgrid
        self.vgrid = Vgrid.default() if vgrid is None else vgrid

    def fetch_data(self, outdir: Union[str, os.PathLike], start_date, rnday, ocean_bnd_ids = [0], elev2D=True, TS=True, UV=True, restart=False, adjust2D=False, lats=None, msl_shifts=None): 
        outdir = pathlib.Path(outdir)

        self.start_date = start_date
        self.rnday=rnday
        self.timevector=np.arange(
            self.start_date,
            self.start_date + timedelta(days=self.rnday+1),
            timedelta(days=1)).astype(datetime)

        #Get open boundary 
        gdf=self.hgrid.boundaries.open.copy()
        opbd=[]
        #for boundary in gdf.itertuples():
        #    opbd.extend(list(boundary.indexes))
        for ibnd in ocean_bnd_ids:
            opbd.extend(list(gdf.iloc[ibnd].indexes))
        blon = self.hgrid.coords[opbd,0]
        blat = self.hgrid.coords[opbd,1]
        #logger.info(f'blon min {np.min(blon)}, max {np.max(blon)}')
        NOP = len(blon)

        #calculate zcor for 3D
        if TS or UV:
            vd=Vgrid.open(self.vgrid)
            sigma=vd.sigma

            #get bathymetry
            depth = self.hgrid.values

            #compute zcor
            zcor = depth[:,None]*sigma
            nvrt=zcor.shape[1]

            #zcor2=zcor[opbd,:]
            #idxs=np.where(zcor2 > 5000)
            #zcor2[idxs]=5000.0-1.0e-6

            #construct schism grid
            #x2i=np.tile(xi,[nvrt,1]).T
            #y2i=np.tile(yi,[nvrt,1]).T
            #bxyz=np.c_[zcor2.reshape(np.size(zcor2)),y2i.reshape(np.size(y2i)),x2i.reshape(np.size(x2i))]
            #logger.info('Computing SCHISM zcor is done!')

        #create netcdf
        ntimes=self.rnday+1
        nComp1=1
        nComp2=2
        one=1
        #ndt=np.zeros([ntimes])

        if elev2D and restart == False:
            #timeseries_el=np.zeros([ntimes,NOP,nComp1])
            #create netcdf 
            dst_elev = Dataset(outdir / 'elev2D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_elev.createDimension('nOpenBndNodes', NOP)
            dst_elev.createDimension('one', one)
            dst_elev.createDimension('time', None)
            dst_elev.createDimension('nLevels', one)
            dst_elev.createDimension('nComponents', nComp1)

            #variables
            dst_elev.createVariable('time_step', 'f', ('one',))
            dst_elev['time_step'][:] = 86400

            dst_elev.createVariable('time', 'f', ('time',))
            #dst_elev['time'][:] = ndt

            dst_elev.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_elev['time_series'][:,:,:,:] = timeseries_el
        elif elev2D and restart:
            dst_elev = Dataset(outdir / 'elev2D.th.nc', 'a', format='NETCDF4')
            time_idx_restart = dst_elev['time'][:].shape[0]

        if TS and restart == False:
            #timeseries_s=np.zeros([ntimes,NOP,nvrt,nComp1])
            dst_salt = Dataset(outdir / 'SAL_3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_salt.createDimension('nOpenBndNodes', NOP)
            dst_salt.createDimension('one', one)
            dst_salt.createDimension('time', None)
            dst_salt.createDimension('nLevels', nvrt)
            dst_salt.createDimension('nComponents', nComp1)
            #variables
            dst_salt.createVariable('time_step', 'f', ('one',))
            dst_salt['time_step'][:] = 86400

            dst_salt.createVariable('time', 'f', ('time',))
            #dst_salt['time'][:] = ndt

            dst_salt.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))

            #temp
            #timeseries_t=np.zeros([ntimes,NOP,nvrt,nComp1])

            dst_temp = Dataset(outdir / 'TEM_3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_temp.createDimension('nOpenBndNodes', NOP)
            dst_temp.createDimension('one', one)
            dst_temp.createDimension('time', None)
            dst_temp.createDimension('nLevels', nvrt)
            dst_temp.createDimension('nComponents', nComp1)
            #variables
            dst_temp.createVariable('time_step', 'f', ('one',))
            dst_temp['time_step'][:] = 86400

            dst_temp.createVariable('time', 'f', ('time',))
            #dst_temp['time'][:] = ndt

            dst_temp.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_temp['time_series'][:,:,:,:] = timeseries_t
        elif TS and restart:
            dst_salt = Dataset(outdir / 'SAL_3D.th.nc', 'a', format='NETCDF4')
            dst_temp = Dataset(outdir / 'TEM_3D.th.nc', 'a', format='NETCDF4')
            time_idx_restart = dst_salt['time'][:].shape[0]

        if UV and restart == False:
            #timeseries_uv=np.zeros([ntimes,NOP,nvrt,nComp2])
            dst_uv = Dataset(outdir / 'uv3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_uv.createDimension('nOpenBndNodes', NOP)
            dst_uv.createDimension('one', one)
            dst_uv.createDimension('time', None)
            dst_uv.createDimension('nLevels', nvrt)
            dst_uv.createDimension('nComponents', nComp2)
            #variables
            dst_uv.createVariable('time_step', 'f', ('one',))
            dst_uv['time_step'][:] = 86400

            dst_uv.createVariable('time', 'f', ('time',))
            #dst_uv['time'][:] = ndt
            
            dst_uv.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_uv['time_series'][:,:,:,:] = timeseries_uv

        elif UV and restart:
            dst_uv = Dataset(outdir / 'uv3D.th.nc', 'a', format='NETCDF4')
            time_idx_restart = dst_uv['time'][:].shape[0]

        logger.info('**** Accessing GOFS data*****')
        t0=time()

        if restart == False: 
            timevector = self.timevector
            it0 = 0
        elif restart:
            #restart from one day earlier
            timevector = self.timevector[time_idx_restart-1:]
            it0 = time_idx_restart-1

        for it1, date in enumerate(timevector):
            
            it = it0 + it1
        
            database=get_database(date)
            logger.info(f'Fetching data for {date} from database {database}')

            #loop over each open boundary
            ind1 = 0
            ind2 = 0
            #for boundary in gdf.itertuples():
            for ibnd in ocean_bnd_ids:
                #opbd = list(boundary.indexes)
                opbd = list(gdf.iloc[ibnd].indexes)
                ind1 = ind2
                ind2 = ind1 + len(opbd)
                #logger.info(f'ind1 = {ind1}, ind2 = {ind2}')
                blon = self.hgrid.coords[opbd,0]
                blat = self.hgrid.coords[opbd,1]
                blonc = blon.mean()
                blatc = blat.mean()
                #logger.info(f'blonc = {blon.mean()}, blatc = {blat.mean()}')
                xi,yi = transform_ll_to_cpp(blon, blat, blonc, blatc)
                bxy = np.c_[yi, xi]

                if TS or UV:
                    zcor2=zcor[opbd,:]
                    idxs=np.where(zcor2 > 5000)
                    zcor2[idxs]=5000.0-1.0e-6

                    #construct schism grid
                    x2i=np.tile(xi,[nvrt,1]).T
                    y2i=np.tile(yi,[nvrt,1]).T
                    bxyz=np.c_[zcor2.reshape(np.size(zcor2)),y2i.reshape(np.size(y2i)),x2i.reshape(np.size(x2i))]

                xmin, xmax = np.min(blon), np.max(blon)
                ymin, ymax = np.min(blat), np.max(blat)
                bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)

                time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2, _ = get_idxs(date, database, bbox, lonc=blonc, latc=blatc)

                if date >= datetime.utcnow():
                    date2 = datetime.utcnow() - timedelta(days=1)
                    url = f'https://tds.hycom.org/thredds/dodsC/{database}/FMRC/runs/GLBy0.08_930_FMRC_RUN_' + \
                        f'{date2.strftime("%Y-%m-%dT12:00:00Z")}?depth[0:1:-1],lat[{lat_idx1}:1:{lat_idx2}],' + \
                        f'lon[{lon_idx1}:1:{lon_idx2}],time[{time_idx}],' + \
                        f'surf_el[{time_idx}][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_temp[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'salinity[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_u[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_v[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}]'
                        
                else:
                    url=f'https://tds.hycom.org/thredds/dodsC/{database}?lat[{lat_idx1}:1:{lat_idx2}],' + \
                        f'lon[{lon_idx1}:1:{lon_idx2}],depth[0:1:-1],time[{time_idx}],' + \
                        f'surf_el[{time_idx}][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_temp[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'salinity[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_u[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'water_v[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}]'
                #logger.info(url)
                 
                ds=Dataset(url)
                dep=ds['depth'][:]

                logger.info(f'****Interpolation starts for boundary {ibnd}****')

                #ndt[it]=it*24*3600.

                if elev2D:
                    #ssh
                    ssh=np.squeeze(ds['surf_el'][:,:])

                    ssh_int = interp_to_points_2d(y2, x2, bxy, ssh)
                    dst_elev['time'][it] = it*24*3600.
                    if adjust2D:
                        elev_adjust = np.interp(blat, lats, msl_shifts)
                        dst_elev['time_series'][it,ind1:ind2,0,0] = ssh_int + elev_adjust
                    else:
                        dst_elev['time_series'][it,ind1:ind2,0,0] = ssh_int 

                if TS:
                    #salt
                    salt = np.squeeze(ds['salinity'][:,:,:])

                    salt_int = interp_to_points_3d(dep, y2, x2, bxyz, salt)
                    salt_int = salt_int.reshape(zcor2.shape)
                    #timeseries_s[it,:,:,0]=salt_int
                    dst_salt['time'][it] = it*24*3600.
                    dst_salt['time_series'][it,ind1:ind2,:,0] = salt_int

                    #temp
                    temp = np.squeeze(ds['water_temp'][:,:,:])

                    #Convert temp to potential temp
                    ptemp = ConvertTemp(salt, temp, dep)

                    temp_int = interp_to_points_3d(dep, y2, x2, bxyz, ptemp)
                    temp_int = temp_int.reshape(zcor2.shape)
                    #timeseries_t[it,:,:,0]=temp_int
                    dst_temp['time'][it] = it*24*3600.
                    dst_temp['time_series'][it,ind1:ind2,:,0] = temp_int

                if UV:
                    uvel=np.squeeze(ds['water_u'][:,:,:])
                    vvel=np.squeeze(ds['water_v'][:,:,:])

                    dst_uv['time'][it] = it*24*3600.
                    #uvel
                    uvel_int = interp_to_points_3d(dep, y2, x2, bxyz, uvel)
                    uvel_int = uvel_int.reshape(zcor2.shape)
                    dst_uv['time_series'][it,ind1:ind2,:,0] = uvel_int

                    #vvel
                    vvel_int = interp_to_points_3d(dep, y2, x2, bxyz, vvel)
                    vvel_int = vvel_int.reshape(zcor2.shape)
                    dst_uv['time_series'][it,ind1:ind2,:,1] = vvel_int
                    #timeseries_uv[it,:,:,1]=vvel_int

                ds.close()
        logger.info(f'Writing *th.nc takes {time()-t0} seconds')

class Nudge:

    def __init__(self, hgrid=None, ocean_bnd_ids=None):

        if hgrid is None:
            raise ValueError('No hgrid information!') 
        else:
            self.hgrid = hgrid

        
        if ocean_bnd_ids is None:
            raise ValueError('Please specify indexes for ocean boundaries!') 
        else:
            self.ocean_bnd_ids = ocean_bnd_ids


    def gen_nudge(self, outdir: Union[str, os.PathLike], rlmax = 1.5, rnu_day=0.25):
        """
        set up nudge zone within rlmax distance from the ocean boundary;
        modify the nudging zone width rlmax.
        rlmax can be a uniform value, e.g., rlmax = 1.5 (degree if hgrid is lon/lat)
        """
          
        outdir = pathlib.Path(outdir)

        rnu_max = 1.0 / rnu_day / 86400.0

        #get nudge zone
        lon = self.hgrid.coords[:,0]
        lat = self.hgrid.coords[:,1]
        gdf = self.hgrid.boundaries.open.copy()
        elnode = self.hgrid.elements.array
        NE, NP = elnode.shape[0],len(lon)
        nudge_coeff = np.zeros(NP, dtype=float)

        global_idxs = {}

        t0 = time()
        nudge_coeff = np.zeros(NP, dtype=float)
        for i in self.ocean_bnd_ids:
            print(f'boundary {i}')
            bnd_idxs = gdf.iloc[i].indexes

            dis = abs((lon + 1j*lat)[:, None] - (lon[bnd_idxs] + 1j*lat[bnd_idxs])[None, :]).min(axis=1)
            out = (1-dis/rlmax)*rnu_max
            out[out<0] = 0
            out[out>rnu_max] = rnu_max
            fp = out>0
            nudge_coeff[fp] = np.maximum(out[fp], nudge_coeff[fp])

            idxs_nudge=np.zeros(NP, dtype=int)
            idxs=np.where(out > 0)[0]
            idxs_nudge[idxs]=1
  
            #expand nudging marker to neighbor nodes
            i34 = self.hgrid.elements.i34
            fp = i34==3
            idxs=np.where(np.max(out[elnode[fp, 0:3]], axis=1) > 0)[0]
            idxs_nudge[elnode[fp,0:3][idxs,:]]=1
            idxs=np.where(np.max(out[elnode[~fp, :]], axis=1) > 0)[0]
            idxs_nudge[elnode[~fp,:][idxs,:]]=1

            idxs=np.where(idxs_nudge == 1)[0]
            global_idxs[i] = idxs


        #logger.info(f'len of nudge idxs is {len(idxs)}')
        logger.info(f'It took {time() -t0} sencods to calcuate nudge coefficient')

        nudge = [f"rlmax={rlmax}, rnu_day={rnu_day}"]
        nudge.extend("\n")
        nudge.append(f"{NE} {NP}")
        nudge.extend("\n")
        hgrid = self.hgrid.to_dict()
        nodes = hgrid['nodes']
        elements = hgrid['elements']
        for idn, (coords, values) in nodes.items():
            line = [f"{idn}"]
            line.extend([f"{x:<.7e}" for x in coords])
            line.extend([f"{nudge_coeff[int(idn)-1]:<.7e}"])
            line.extend("\n")
            nudge.append(" ".join(line))

        for id, element in elements.items():
            line = [f"{id}"]
            line.append(f"{len(element)}")
            line.extend([f"{e}" for e in element])
            line.extend("\n")
            nudge.append(" ".join(line))

        with open(outdir / 'TEM_nudge.gr3','w+') as fid:
            fid.writelines(nudge)

        shutil.copy2(outdir / 'TEM_nudge.gr3', outdir / 'SAL_nudge.gr3')

        return global_idxs

    def fetch_data(self, outdir: Union[str, os.PathLike], vgrid, start_date, rnday, restart = False, rlmax = None, rnu_day=None):
        """
        fetch data from the database and generate nudge file
        see gen_nudge for the meaning of rlmax, rnu_day
        """

        outdir = pathlib.Path(outdir)

        self.start_date = start_date
        self.rnday=rnday
        self.timevector=np.arange(
            self.start_date,
            self.start_date + timedelta(days=self.rnday+1),
            timedelta(days=1)).astype(datetime)

        vd=Vgrid.open(vgrid)
        sigma=vd.sigma

        #define nudge zone and strength 
        rlmax = 1.5 if rlmax is None else rlmax
        rnu_day = 0.25 if rnu_day is None else rnu_day
        logger.info(f'Max relax distance is {rlmax} degree, max relax strengh is {rnu_day} days.')
        #Get the index for nudge
        global_idxs = self.gen_nudge(outdir, rlmax = rlmax, rnu_day=rnu_day)

        #get bathymetry
        depth = self.hgrid.values

        #compute zcor
        zcor = depth[:,None]*sigma
        nvrt=zcor.shape[1]

        #allocate output variables
        include = global_idxs[0]
        nbnd = len(self.ocean_bnd_ids)
        if nbnd > 1:
            include = np.concatenate((global_idxs[0], global_idxs[1],))
            if nbnd > 2:
                for i in self.ocean_bnd_ids[2:]:
                    include = np.concatenate((include, global_idxs[i]))
        else:
            include = global_idxs[0]

        nNode = include.shape[0]
        one = 1
        ntimes = self.rnday+1

        #timeseries_s=np.zeros([ntimes,nNode,nvrt,one])
        #timeseries_t=np.zeros([ntimes,nNode,nvrt,one])
        #ndt=np.zeros([ntimes])
        if restart:
            dst_temp = Dataset(outdir / 'TEM_nu.nc', 'a', format='NETCDF4')
            dst_salt = Dataset(outdir / 'SAL_nu.nc', 'a', format='NETCDF4')
            time_idx_restart = dst_temp['time'][:].shape[0]
        else:
            dst_temp = Dataset(outdir / 'TEM_nu.nc', 'w', format='NETCDF4')
            #dimensions
            dst_temp.createDimension('node', nNode)
            dst_temp.createDimension('nLevels', nvrt)
            dst_temp.createDimension('one', one)
            dst_temp.createDimension('time', None)
            #variables
            dst_temp.createVariable('time', 'f', ('time',))
            #dst_temp['time'][:] = ndt

            dst_temp.createVariable('map_to_global_node', 'i4', ('node',))
            dst_temp['map_to_global_node'][:] = include+1

            dst_temp.createVariable('tracer_concentration', 'f', ('time', 'node', 'nLevels', 'one'))
            #dst_temp['tracer_concentration'][:,:,:,:] = timeseries_t

            #salinity
            dst_salt = Dataset(outdir / 'SAL_nu.nc', 'w', format='NETCDF4')
            #dimensions
            dst_salt.createDimension('node', nNode)
            dst_salt.createDimension('nLevels', nvrt)
            dst_salt.createDimension('one', one)
            dst_salt.createDimension('time', None)
            #variables
            dst_salt.createVariable('time', 'f', ('time',))
            #dst_salt['time'][:] = ndt

            dst_salt.createVariable('map_to_global_node', 'i4', ('node',))
            dst_salt['map_to_global_node'][:] = include+1

            dst_salt.createVariable('tracer_concentration', 'f', ('time', 'node', 'nLevels', 'one'))
            #dst_salt['tracer_concentration'][:,:,:,:] = timeseries_s

        logger.info('**** Accessing GOFS data*****')
        if restart:
            #restart from one day earlier to make sure all files consistant
            timevector = self.timevector[time_idx_restart-1:]
            it0 = time_idx_restart-1
        else:
            timevector = self.timevector
            it0 = 0

        t0=time()
        for it1, date in enumerate(timevector):

            it = it0 + it1

            database=get_database(date)
            logger.info(f'Fetching data for {date} from database {database}')

            ind1 = 0
            ind2 = 0
            for ibnd in self.ocean_bnd_ids:
                include = global_idxs[ibnd]

                ind1 = ind2
                ind2 = ind1 + include.shape[0]
                #Get open nudge array 
                nlon = self.hgrid.coords[include, 0]
                nlat = self.hgrid.coords[include, 1]
                nlonc = nlon.mean()
                nlatc = nlat.mean()
                xi,yi = transform_ll_to_cpp(nlon, nlat, nlonc, nlatc)
                bxy = np.c_[yi, xi]

                zcor2=zcor[include,:]
                idxs=np.where(zcor2 > 5000)
                zcor2[idxs]=5000.0-1.0e-6

                #construct schism grid
                x2i=np.tile(xi,[nvrt,1]).T
                y2i=np.tile(yi,[nvrt,1]).T
                bxyz=np.c_[zcor2.reshape(np.size(zcor2)),y2i.reshape(np.size(y2i)),x2i.reshape(np.size(x2i))]
                logger.info('Computing SCHISM zcor is done!')

                xmin, xmax = np.min(nlon), np.max(nlon)
                ymin, ymax = np.min(nlat), np.max(nlat)
                bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)
                logger.info(f'bbox for nudge is {bbox}')

                time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2, _ = get_idxs(date, database, bbox, lonc=nlonc, latc=nlatc)

                if date >= datetime.utcnow():
                    date2 = datetime.utcnow() - timedelta(days=1)
                    url = f'https://tds.hycom.org/thredds/dodsC/{database}/FMRC/runs/GLBy0.08_930_FMRC_RUN_' + \
                        f'{date2.strftime("%Y-%m-%dT12:00:00Z")}?depth[0:1:-1],lat[{lat_idx1}:1:{lat_idx2}],' + \
                        f'lon[{lon_idx1}:1:{lon_idx2}],time[{time_idx}],' + \
                        f'water_temp[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'salinity[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}]'

                else:
                    url=f'https://tds.hycom.org/thredds/dodsC/{database}?lat[{lat_idx1}:1:{lat_idx2}],' + \
                        f'lon[{lon_idx1}:1:{lon_idx2}],depth[0:1:-1],time[{time_idx}],' + \
                        f'water_temp[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                        f'salinity[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}]'

                ds=Dataset(url)
                salt=np.squeeze(ds['salinity'][:,:,:])
                temp=np.squeeze(ds['water_temp'][:,:,:])

                #Convert temp to potential temp
                dep=ds['depth'][:]
                ptemp = ConvertTemp(salt, temp, dep)

                logger.info('****Interpolation starts****')

                #ndt[it]=it
                #salt
                dst_salt['time'][it] = it
                salt_int = interp_to_points_3d(dep, y2, x2, bxyz, salt)
                salt_int = salt_int.reshape(zcor2.shape)
                #timeseries_s[it,:,:,0]=salt_int
                dst_salt['tracer_concentration'][it,ind1:ind2,:,0] = salt_int

                #temp
                dst_temp['time'][it] = it
                temp_int = interp_to_points_3d(dep, y2, x2, bxyz, ptemp)
                temp_int = temp_int.reshape(zcor2.shape)
                #timeseries_t[it,:,:,0]=temp_int
                dst_temp['tracer_concentration'][it,ind1:ind2,:,0] = temp_int

                ds.close()
 
        #dst_temp.close()
        #dst_salt.close()
        
        logger.info(f'Writing *_nu.nc takes {time()-t0} seconds')

class DownloadHycom:

    def __init__(self, hgrid=None, bbox=None):

        if hgrid is None and bbox is None:
            raise ValueError('Either hgrid or bbox must be provided!') 

        if hgrid is not None:        
            xmin, xmax = hgrid.coords[:, 0].min(), hgrid.coords[:, 0].max()
            ymin, ymax = hgrid.coords[:, 1].min(), hgrid.coords[:, 1].max()
            self.bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)
        elif bbox is not None:
            self.bbox = bbox

    def fetch_data(self, start_date, rnday=1, fmt='schism', bnd=False, nudge=False, sub_sample=1, outdir=None):
        '''
        start_date: datetime.datetime
        rnday: integer
        fmt: 'schism' - for Fortran code; 'hycom' - raw netCDF from HYCOM
        bnd: file names are SSH_*.nc, TS_*.nc, UV_*.nc used in gen_hot_3Dth_from_hycom.f90
        nudge: file name is TS_*.nc used in gen_nudge_from_hycom.f90
        outdir: directory for output files
        '''
        if rnday == 1:
            timevector = [start_date]
        else:
            timevector = np.arange(
                start_date, start_date + timedelta(days=rnday+1), timedelta(days=1)
            ).astype(datetime)

        for i, date in enumerate(timevector):
            database=get_database(date)
            logger.info(f'Fetching data for {date} from database {database}')

            time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2, isLonSame = get_idxs(date, database, self.bbox)

            if fmt == 'schism':
                url_ssh = f'https://tds.hycom.org/thredds/dodsC/{database}?lat[{lat_idx1}:{sub_sample}:{lat_idx2}],' + \
                    f'lon[{lon_idx1}:{sub_sample}:{lon_idx2}],depth[0:1:-1],time[{time_idx}],' + \
                    f'surf_el[{time_idx}][{lat_idx1}:{sub_sample}:{lat_idx2}][{lon_idx1}:{sub_sample}:{lon_idx2}],' + \
                    f'water_u[{time_idx}][0:1:39][{lat_idx1}:{sub_sample}:{lat_idx2}][{lon_idx1}:{sub_sample}:{lon_idx2}],' + \
                    f'water_v[{time_idx}][0:1:39][{lat_idx1}:{sub_sample}:{lat_idx2}][{lon_idx1}:{sub_sample}:{lon_idx2}],' + \
                    f'water_temp[{time_idx}][0:1:39][{lat_idx1}:{sub_sample}:{lat_idx2}][{lon_idx1}:{sub_sample}:{lon_idx2}],' + \
                    f'salinity[{time_idx}][0:1:39][{lat_idx1}:{sub_sample}:{lat_idx2}][{lon_idx1}:{sub_sample}:{lon_idx2}]'

                foutname = f'hycom_{date.strftime("%Y%m%d")}.nc'
                #foutname = f'TS_{i+1}.nc'
                logger.info(f'filename is {foutname}')
                ds = xr.open_dataset(url_ssh)

                #convert in-situ temperature to potential temperature
                temp = ds.water_temp.values
                salt = ds.salinity.values
                dep = ds.depth.values

                ptemp = ConvertTemp(salt, temp, dep)
                #drop water_temp variable and add new temperature variable
                ds = ds.drop('water_temp')
                ds['temperature']=(['time','depth','lat','lon'], ptemp)
                ds.temperature.attrs = {
                    'long_name': 'Sea water potential temperature',
                    'standard_name': 'sea_water_potential_temperature',
                    'units': 'degC'
                }

                if not isLonSame:
                    logger.info('Lon is not the same!')
                    ds = convert_longitude(ds, self.bbox)

                ds = ds.rename_dims({'lon':'xlon'})
                ds = ds.rename_dims({'lat':'ylat'})
                ds = ds.rename_vars({'lat':'ylat'})
                ds = ds.rename_vars({'lon':'xlon'})

                t0 =  time()
                logger.info(f'Start writing nc file!')
                ds.to_netcdf(foutname, 'w', unlimited_dims='time', encoding={'temperature':{'dtype': 'h', '_FillValue': -30000.,'scale_factor': 0.001, 'add_offset': 20., 'missing_value': -30000.}})
                ds.close()
                logger.info(f'It took {time()-t0} seconds to write nc file!')

                #link output file to Fortran required files
                dir_path = os.path.abspath(outdir)
                logger.info(f'current dir is {dir_path}')
                src = f'{dir_path}/{foutname}'
                if bnd:
                    names = ['SSH', 'TS', 'UV']
                    for name in names:
                        dst = f'{dir_path}/{name}_{i+1}.nc'
                        os.symlink(src, dst)
                elif nudge:
                        dst = f'{dir_path}/TS_{i+1}.nc'
                        os.symlink(src, dst)

            elif fmt == 'hycom':
                url=f'https://tds.hycom.org/thredds/dodsC/{database}?lat[{lat_idx1}:1:{lat_idx2}],' + \
                    f'lon[{lon_idx1}:1:{lon_idx2}],depth[0:1:-1],time[{time_idx}],' + \
                    f'surf_el[{time_idx}][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                    f'water_temp[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                    f'salinity[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                    f'water_u[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}],' + \
                    f'water_v[{time_idx}][0:1:39][{lat_idx1}:1:{lat_idx2}][{lon_idx1}:1:{lon_idx2}]'

                foutname = f'hycom_{date.strftime("%Y%m%d")}.nc' 
                ds = xr.open_dataset(url)
                ds.to_netcdf(foutname, 'w')

                ds.close()



from pyschism.mesh.hgrid import Hgrid
from pyschism.mesh.gridgr3 import ElevIc

if __name__ == '__main__':

    hgrid = Hgrid.open('./hgrid.gr3', crs='epsg:4326') 

    elevic = ElevIc.default(hgrid=hgrid, offset=0.1)
    elevic.write('elev.ic', overwrite=True)



import argparse
from datetime import timedelta
import pathlib
import sys

import matplotlib.pyplot as plt

from pyschism.cmd import common
from pyschism.cmd.base import CliComponent
from pyschism.enums import Sflux1Types, Sflux2Types
from pyschism.forcing.nws.nws2 import NWS2
from pyschism.mesh import gridgr3

sflux_variables = {
    "air": ["prmsl", "spfh", "stmp", "uwind", "vwind"],
    "prc": ["prate"],
    "rad": ["dlwrf", "dswrf"],
}


class AnimateSubCli:
    def __init__(self, args):
        kwargs = {}

        if args.sflux_1_glob is not None:
            kwargs.setdefault("sflux_1_glob", args.sflux_1_glob)

        if args.sflux_2_glob is not None:
            kwargs.setdefault("sflux_2_glob", args.sflux_2_glob)

        if pathlib.Path(args.sflux_directory).is_dir() is False:

            class NotADirectory(Exception):
                pass

            raise NotADirectory(f"{args.sflux_directory} is not a directory.")

        nws2 = NWS2.read(args.sflux_directory, **kwargs)

        sflux = getattr(nws2, f"sflux_{args.level}")
        for vargroup, vars in sflux_variables.items():
            component = getattr(sflux, f"{vargroup}")
            for variable in vars:
                animations = []
                if getattr(args, variable) is True:
                    animations.append(
                        getattr(component, f"{variable}").animation(
                            save=args.save / f"{variable}.gif",
                            fps=args.fps,
                        )
                    )

            if args.show is True:
                plt.show()


class GenerateSubCli:
    def __init__(self, args):
        if args.file_interval_sflux_1 is not None:
            args.sflux.sflux_1.inventory.file_interval = args.file_interval_sflux_1
        if args.file_interval_sflux_2 is not None:
            args.sflux.sflux_2.inventory.file_interval = args.file_interval_sflux_2
        args.sflux.write(
            args.output_directory,
            start_date=args.start_date,
            end_date=args.end_date,
            bbox=args.hgrid.get_bbox(crs="epsg:4326"),
            overwrite=args.overwrite,
            windrot=args.windrot,
            air=args.air,
            rad=args.rad,
            prc=args.prc,
        )


class SfluxCli(CliComponent):
    def __init__(self, args: argparse.Namespace):
        exec(f"{args.action.capitalize()}SubCli(args)")

    @staticmethod
    def add_subparser_action(subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser("sflux")
        subparsers = parser.add_subparsers(dest="action")
        add_animate_to_subparsers(subparsers)
        add_generate_to_subparsers(subparsers)


def add_animate_to_subparsers(subparsers):
    parser = subparsers.add_parser("animate")
    parser.add_argument("sflux_directory")
    parser.add_argument("--level", choices=["1", "2"], default="1")
    parser.add_argument("--sflux-1-glob")
    parser.add_argument("--sflux-2-glob")
    for vargroup, variables in sflux_variables.items():
        arg_group = parser.add_argument_group(f"{vargroup} variables")
        for variable in variables:
            arg_group.add_argument(f"--{variable}", action="store_true", default=False)
    parser.add_argument("--save", type=lambda x: pathlib.Path(x))
    parser.add_argument("--fps", type=float)
    parser.add_argument(
        "--show",
        action="store_true",
        dest="show",
    )
    parser.add_argument(
        "--no-show",
        action="store_true",
        dest="show",
    )
    parser.set_defaults(show=False if "--save" in " ".join(sys.argv) else True)


def add_generate_to_subparsers(subparsers):
    parser = subparsers.add_parser("generate")
    # add_sflux_to_parser(parser)
    common.add_hgrid_to_parser(parser)
    # add_sflux_to_parser(parser)
    common.add_sflux_to_parser(parser)
    common.add_dates_to_parser(parser)
    parser.add_argument("--output-directory", "-o", required=True, type=pathlib.Path)
    parser.add_argument("--overwrite", action="store_true", default=False)
    air = parser.add_argument_group("air group").add_mutually_exclusive_group()
    air.add_argument("--air", action="store_true", dest="air")
    air.add_argument("--no-air", action="store_false", dest="air")
    prc = parser.add_argument_group("prc group").add_mutually_exclusive_group()
    prc.add_argument("--prc", action="store_true", dest="prc")
    prc.add_argument("--no-prc", action="store_false", dest="prc")
    rad = parser.add_argument_group("rad group").add_mutually_exclusive_group()
    rad.add_argument("--rad", action="store_true", dest="rad")
    rad.add_argument("--no-rad", action="store_false", dest="rad")
    parser.set_defaults(
        air=True,
        prc=True,
        rad=True,
    )
    common.add_log_level_to_parser(parser)


# def add_sflux_to_parser(parser):

#     class SfluxAction(argparse.Action):

#         def __call__(self, parser, namespace, values, option_string=None):
#             if len(values) > 2:
#                 raise ValueError(
#                     "pyschism sflux generate: error: argument sflux_databases: "
#                     "expected at most two arguments."
#                 )
#             sflux_1 = Sflux1Types[values[0].upper()].value(
#                 product="gfs_0p25_1hr" if values[0].lower() == "gfs" else values[0].lower()
#             )
#             if len(values) == 2:
#                 sflux_2 = Sflux2Types[values[1].upper()].value()
#             else:
#                 sflux_2 = None

#             # tmp_parser = argparse.ArgumentParser(add_help=False)
#             # if not bool(set(sys.argv).intersection(['-h', '--help'])):
#             #     common.add_windrot_to_parser(tmp_parser)
#             #     tmp_args = tmp_parser.parse_known_args()[0]
#             #     windrot = gridgr3.Windrot.default(namespace.hgrid) if tmp_args.windrot is None else tmp_args.windrot
#             # else:
#             #     windrot = None
#             setattr(
#                 namespace,
#                 self.dest,
#                 NWS2(
#                     sflux_1,
#                     sflux_2,
#                     windrot=windrot
#                 )
#             )

#     parser.add_argument(
#         'sflux_databases',
#         nargs="+",
#         action=SfluxAction,
#         metavar="sflux_database",
#         help='Names of the two sflux levels. At least one name is required. Valid choices are:'
#              '`gfs`, `hrrr`.',
#     )



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from functools import lru_cache
import logging
from typing import Dict, Union

from matplotlib.transforms import Bbox
import seawater as sw
from netCDF4 import Dataset
import numpy as np
import requests
from scipy.interpolate import griddata, RegularGridInterpolator, NearestNDInterpolator
import tqdm
import tqdm_logging_wrapper
import xmltodict

from pyschism import dates
from pyschism.forcing.hycom.base import Hycom, HycomComponent

logger = logging.getLogger(__name__)


class GofsDatasetCollection(ABC):

    start_date = dates.StartDate()
    end_date = dates.EndDate()

    def __init__(
            self,
            start_date: datetime = None,
            end_date: Union[datetime, timedelta] = None,
            output_interval: timedelta = None
    ):
        self.start_date = start_date
        self.end_date = self.maximum_end_date - self.sampling_interval if \
            end_date is None else end_date 
        self.output_interval = self.sampling_interval if output_interval is None \
            else output_interval
        if (self.output_interval % self.sampling_interval).total_seconds() != 0:
            raise ValueError(
                'Argument output_interval must be modulus of '
                f'{self.sampling_interval} hours.')

    @property
    @abstractmethod
    def datasets(self):
        '''Get the datasets based on the set start_date and end_date dates.'''

    @property
    def sampling_frequency(self):
        return 1./self.sampling_interval.total_seconds()

    @property
    def sampling_interval(self):
        return timedelta(hours=3)

    @property
    def required_datevector(self):
        required_start_date = dates.nearest_cycle(self.start_date)
        required_end_date = dates.nearest_cycle(
            self.end_date + self.output_interval,
            period=int(self.output_interval/timedelta(hours=1))
        )
        return np.arange(
            required_start_date,
            required_end_date + self.output_interval,
            self.output_interval
        ).astype(datetime)


class GofsForecastDatasets(GofsDatasetCollection):

    catalog_url = 'https://tds.hycom.org/thredds/catalog/GLBy0.08/expt_93.0/FMRC/runs/catalog.xml'
    base_url = 'https://tds.hycom.org/thredds/dodsC/'

    @property
    def datasets(self):
        datasets = {}
        for required_date in self.required_datevector:
            for i, datevector in enumerate(self.datevectors):
                if required_date in datevector:
                    opendap_url = self.base_url + self.xmlcatalog[
                        'catalog']['dataset']['dataset'][i]['@urlPath']
                    datasets.setdefault(
                        required_date,
                        Dataset(opendap_url)
                    )
                    break
        return datasets
        
    @property
    def datevectors(self):
        for file in self.xmlcatalog['catalog']['dataset']['dataset']:
            start = datetime.strptime(
                file['timeCoverage']['start'], '%Y-%m-%dT%H:%M:%SZ')
            end = datetime.strptime(
                file['timeCoverage']['end'], '%Y-%m-%dT%H:%M:%SZ')
            steps = int(
                (end - start).total_seconds() * self.sampling_frequency) + 1
            datevec = [start + x*timedelta(
                seconds=1/self.sampling_frequency) for x in range(steps)]
            yield datevec

    @property
    def xmlcatalog(self):
        return xmltodict.parse(requests.get(self.catalog_url).content)

    @property
    def minimum_datetime(self):
        start_dates = []
        for rec in self.xmlcatalog['catalog']['dataset']['dataset']:
            start_dates.append(dates.localize_datetime(datetime.strptime(
                rec['timeCoverage']['start'], '%Y-%m-%dT%H:%M:%SZ')))
        return np.min(start_dates)

    @property
    def maximum_datetime(self):
        end_dates = []
        for rec in self.xmlcatalog['catalog']['dataset']['dataset']:
            end_dates.append(dates.localize_datetime(datetime.strptime(
                rec['timeCoverage']['end'], '%Y-%m-%dT%H:%M:%SZ')))
        return np.max(end_dates)

    @property
    def maximum_time_range(self):
        return self.maximum_datetime - self.minimum_datetime

def get_database(date, Bbox=None):
    if date >= datetime(2018, 12, 4):
        database = f'GLBy0.08/expt_93.0'
    elif date >= datetime(2018, 1, 1) and date < datetime(2018, 12, 4):
        database = f'GLBv0.08/expt_93.0'
    elif date >= datetime(2017, 10, 1) and date < datetime(2018, 1, 1):
        database = f'GLBv0.08/expt_92.9'
    elif date >= datetime(2017, 6, 1) and date < datetime(2017, 10, 1):
        database = f'GLBv0.08/expt_57.7'
    elif date >= datetime(2017, 2, 1) and date < datetime(2017, 6, 1):
        database = f'GLBv0.08/expt_92.8'
    elif date >= datetime(2016, 5, 1) and date < datetime(2017, 2, 1):
        database = f'GLBv0.08/expt_57.2'
    elif date >= datetime(2016, 1, 1) and date < datetime(2016, 5, 1):
        database = f'GLBv0.08/expt_56.3'
    elif date >= datetime(1994, 1, 1) and date < datetime(2016, 1, 3):
        database = f'GLBv0.08/expt_53.X/data/{date.year}'
    else:
        print('No data for {date}')
    return database

class GofsHindcastDatasets(GofsDatasetCollection):

    @property
    def datasets(self):
        datasets = {}
        for required_date in self.required_datevector:
            database = get_database(required_date)
            print(f'Database for {required_date} is {database}')
            #baseurl = f'https://tds.hycom.org/thredds/dodsC/{database}?lat[0:1:3250],lon[0:1:4499],' + \
            #    f'time[0:1:6127],depth[0:1:39]' 
            #ds=Dataset(baseurl)
            #time1=ds['time']
            #times=nc4.num2date(time1,units=time1.units,only_use_cftime_datetimes=False)
            #time_idx=np.where( required_date == times)[0].item()
             
            opendap_url = f'https://tds.hycom.org/thredds/dodsC/{database}?lat[0:-1],lon[0:-1],' + \
                f'time[0:-1],surf_el[0:-1][0:-1][0:-1],depth[0:-1],' + \
                f'water_temp[0:-1][0:-1][0:-1][0:1:-1],' + \
                f'salinity[0:-1][0:-1][0:-1][0:-1],' + \
                f'water_u[0:-1][0:-1][0:-1][0:-1],' + \
                f'water_v[0:-1][0:-1][0:-1][0:-1]'
            datasets.setdefault(
                required_date,
                Dataset(opendap_url)
                    )
        return datasets
        #raise NotImplementedError('Need to return the datasets.')

        

    def pad_datasets(self, datasets):
        for dataset in datasets.values():
            if dataset is None:
                raise NotImplementedError(
                    'Hindcast data fetching is unavailable.')


class GofsDatasets:

    def __init__(self, start_date, run_days, output_interval):
        print(f'start_date is {start_date}')
        print(f'today is {datetime.now().strftime("%Y-%m-%d")}')
        self.start_date = start_date
        if start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
            self.hindcast = GofsHindcastDatasets(start_date, run_days, output_interval)
        else:
            self.forecast = GofsForecastDatasets(start_date, run_days, output_interval)

    @property
    def datasets(self):
        if self.start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
            datasets = self.hindcast.datasets
        else:
            datasets = self.forecast.datasets
        #self.hindcast.pad_datasets(datasets)
        return datasets


class GOFSComponent(HycomComponent):

    @lru_cache(maxsize=None)
    def get_datasets(
            self,
            start_date: datetime,
            run_days: Union[float, timedelta],
            output_interval=timedelta(days=1)
    ) -> Dict[datetime, Dataset]:
        return GofsDatasets(start_date, run_days, output_interval).datasets

def transform_ll_to_cpp(lon, lat, lonc=-77.07, latc=24.0):
    #lonc=(np.max(lon)+np.min(lon))/2.0
    print(f'lonc is {lonc}')
    #latc=(np.max(lat)+np.min(lat))/2.0
    print(f'latc is {latc}')
    longitude=lon/180*np.pi
    latitude=lat/180*np.pi
    radius=6378206.4
    loncc=lonc/180*np.pi
    latcc=latc/180*np.pi
    lon_new=[radius*(longitude[i]-loncc)*np.cos(latcc) for i in np.arange(len(longitude))]
    lat_new=[radius*latitude[i] for i in np.arange(len(latitude))]

    return np.array(lon_new), np.array(lat_new)

class GOFSElevation(GOFSComponent):

    @property
    def ncvar(self):
        return 'surf_el'

    def put_boundary_ncdata(
            self,
            boundary,
            dst,
            start_date,
            run_days,
            overwrite=False,
            offset=0,
            output_interval=timedelta(hours=24),
            pixel_buffer=10,
            progress_bar=True
    ):
        for i, (time, dataset) in enumerate(
            self.get_datasets(
                start_date,
                run_days,
                output_interval
            ).items()
        ):
            if start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:]), 
                    '%Y-%m-%d%H:%M:%S')
                #time1 = dataset['time']
                #ds_timevector = nc4.num2date(
                #    time1,
                #    units=time1.units,
                #    only_use_cftime_datetimes=False) 
            else:
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:-1]),
                    '%Y-%m-%d%H:%M:%S.%f')
            ds_timevector = [ds_base_date + timedelta(hours=x)
                         for x in dataset['time'][:]]
            requested_date = dates.nearest_cycle(
                start_date + i*output_interval,
                period=3).replace(tzinfo=None)
            time_idx = ds_timevector.index(requested_date)
            logger.info(
                'Saving GOFS elev data for date: '
                f'{start_date+i*output_interval} '
                f'approximated as {ds_timevector[time_idx]} for '
                f'boundary id={boundary.id}'
                )
            bounds = boundary.geometry.bounds
            dx = (dataset['lon'][-1] - dataset['lon'][0]) / len(dataset['lon'])
            dy = (dataset['lat'][-1] - dataset['lat'][0]) / len(dataset['lat'])
            bounds = (
                bounds[0] - 2*dx,
                bounds[1] - 2*dy,
                bounds[2] + 2*dx,
                bounds[3] + 2*dy,
                )
            bbox = self._modified_bbox(
                dataset, Bbox.from_extents(*bounds))
            lon_idxs, lat_idxs = self._modified_bbox_indexes(
                    bbox,
                    dataset,
                    pixel_buffer
                )
            #zi = np.full((len(lat_idxs), len(lon_idxs)), np.nan)
            if progress_bar is True:
                items_iter = tqdm.tqdm(lat_idxs)
                with tqdm_logging_wrapper.wrap_logging_for_tqdm(
                        items_iter), items_iter:
                    #for k, lat_idx in enumerate(items_iter):
                    #    zi[k, :] = dataset[self.ncvar][
                    #                        time_idx, lat_idx, lon_idxs]
                    zi = dataset[self.ncvar][time_idx, lat_idxs, lon_idxs]
            else:
                #for k, lat_idx in enumerate(lat_idxs):
                #    zi[k, :] = dataset[self.ncvar][
                #                        time_idx, lat_idx, lon_idxs]
                zi = dataset[self.ncvar][time_idx, lat_idxs, lon_idxs]
            idxs = np.where(abs(zi) > 10000)
            zi[idxs] = float('nan')

            xi = dataset['lon'][lon_idxs]
            for idx in range(len(xi)):
                if xi[idx] > 180:
                    xi[idx] = xi[idx]-360.
            yi = dataset['lat'][lat_idxs]
            xi, yi = np.meshgrid(xi, yi)
            xi = xi.flatten()
            yi = yi.flatten()
            zi = zi.flatten()
            xyq = np.array(boundary.geometry.coords)
            zq = griddata(
                (xi, yi),
                zi,
                (xyq[:, 0], xyq[:, 1]),
                method='linear',
                fill_value=np.nan,
            )
            nan_idxs = np.where(np.isnan(zq))
            q_non_nan = np.where(~np.isnan(zi))
            zq[nan_idxs] = griddata(
                (xi[q_non_nan], yi[q_non_nan]),
                zi[q_non_nan],
                (xyq[nan_idxs, 0], xyq[nan_idxs, 1]),
                method='nearest',
            )
            if np.any(np.isnan(zq)):
                raise ValueError('Boundary contains NaNs.')
            print(f'the shape of zq is {len(zq)}, max zq is {np.max(zq)}, min zq is {np.min(zq)}')
            dst['time_series'][i, offset:offset+len(zq)] = zq


class GOFSVelocity(GOFSComponent):

    @property
    def ncvar(self):
        return 'water_u', 'water_v'

    def put_boundary_ncdata(
            self,
            hgrid,
            vgrid,
            boundary,
            dst,
            start_date,
            run_days,
            overwrite=False,
            offset=0,
            output_interval=timedelta(hours=24),
            pixel_buffer=10,
            progress_bar=True
    ):

        for i, (time, dataset) in enumerate(
            self.get_datasets(
                start_date,
                run_days,
                output_interval
            ).items()
        ):
            if start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:]),
                    '%Y-%m-%d%H:%M:%S')
                #time1 = dataset['time']
                #ds_timevector = nc4.num2date(
                #    time1,
                #    units=time1.units,
                #    only_use_cftime_datetimes=False) 
            else:
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:-1]),
                    '%Y-%m-%d%H:%M:%S.%f')
            ds_timevector = [ds_base_date + timedelta(hours=x)
                             for x in dataset['time'][:]]
            requested_date = dates.nearest_cycle(
                start_date + i*output_interval,
                period=3).replace(tzinfo=None)
            time_idx = ds_timevector.index(requested_date)
            logger.info(
                'Saving GOFS water_uv data for date: '
                f'{start_date+i*output_interval} '
                f'approximated as {ds_timevector[time_idx]} for '
                f'boundary id={boundary.id}'
                )
            bounds = boundary.geometry.bounds
            dx = (dataset['lon'][-1] - dataset['lon'][0]) / len(dataset['lon'])
            dy = (dataset['lat'][-1] - dataset['lat'][0]) / len(dataset['lat'])
            bounds = (
                bounds[0] - 2*dx,
                bounds[1] - 2*dy,
                bounds[2] + 2*dx,
                bounds[3] + 2*dy,
                )
            bbox = self._modified_bbox(
                dataset, Bbox.from_extents(*bounds))
            lon_idxs, lat_idxs = self._modified_bbox_indexes(
                    bbox,
                    dataset,
                    pixel_buffer
                )
            uvar, vvar = self.ncvar
            # z_ui_idxs = list(range(dataset['depth'].shape[0]))
            z_idxs = list(range(dataset['depth'].shape[0]))  # TODO: subset?
            #ui = np.full((len(z_idxs), len(lat_idxs), len(lon_idxs)), np.nan)
            #vi = np.full((len(z_idxs), len(lat_idxs), len(lon_idxs)), np.nan)
            if progress_bar is True:
                items_iter = tqdm.tqdm(lat_idxs)
                with tqdm_logging_wrapper.wrap_logging_for_tqdm(
                        items_iter), items_iter:
                    #for k, lat_idx in enumerate(items_iter):
                    #    ui[:, k, :] = dataset[uvar][time_idx, z_idxs, lat_idx, lon_idxs]
                    #    vi[:, k, :] = dataset[vvar][time_idx, z_idxs, lat_idx, lon_idxs]
                    ui = dataset[uvar][time_idx, :, lat_idxs, lon_idxs]
                    vi = dataset[vvar][time_idx, :, lat_idxs, lon_idxs]
            else:
                #for k, lat_idx in enumerate(lat_idxs):
                #    ui[:, k, :] = dataset[uvar][time_idx, z_idxs, lat_idx, lon_idxs]
                #    vi[:, k, :] = dataset[vvar][time_idx, z_idxs, lat_idx, lon_idxs]
                ui = dataset[uvar][time_idx, :, lat_idxs, lon_idxs]
                vi = dataset[vvar][time_idx, :, lat_idxs, lon_idxs]
            #change missing value to nan
            idxs = np.where(abs(ui) > 10000)
            ui[idxs] = float('nan')
            idxs = np.where(abs(vi) > 10000)
            vi[idxs] = float('nan')

            loni = dataset['lon'][lon_idxs]
            for idx in range(len(loni)):
                if loni[idx] > 180:
                    loni[idx] = loni[idx]-360.
            lati = dataset['lat'][lat_idxs]
            xi, yi = transform_ll_to_cpp(loni, lati)

            if vgrid.ivcor == 1:
                bz = (hgrid.values[:, None]*vgrid.sigma)[boundary.indexes, :]
                idxs = np.where(bz > 5000.0)
                bz[idxs] = 5000.0 - 1.0e-6 
            else:
                raise NotImplementedError('vgrid.ivcor!=1')

            xy = hgrid.get_xy(crs='epsg:4326')
            lonb = xy[boundary.indexes, 0]
            latb = xy[boundary.indexes, 1]
            xb, yb = transform_ll_to_cpp(lonb, latb)
            bx = np.tile(xb, [bz.shape[1],1]).T
            by = np.tile(yb, [bz.shape[1],1]).T
            bzyx = np.c_[bz.reshape(np.size(bz)), by.reshape(np.size(by)), bx.reshape(np.size(bx))]
            zi = dataset['depth'][z_idxs]

            # First try with RegularGridInterpolator
            ui_fd = RegularGridInterpolator(
                (zi, yi, xi),
                ui,
                method='linear',
                bounds_error=False,
                fill_value=np.nan
            )
            u_interp = ui_fd(bzyx)

            # the boundary and the data don't intersect
            if np.all(np.isnan(u_interp)):
                ui_idxs = np.where(~np.isnan(ui))
                xyzi = np.vstack([
                    np.tile(xi, ui.shape)[ui_idxs].flatten(),
                    np.tile(yi, ui.shape)[ui_idxs].flatten(),
                    np.tile(zi, ui.shape)[ui_idxs].flatten()
                ]).T
                ui_fd = NearestNDInterpolator(xyzi, ui[ui_idxs].flatten())
                u_interp = ui_fd(np.vstack([bx, by, -bz.flatten()]).T)
            # the boundary and the data partially intersect
            elif np.any(np.isnan(u_interp)):
                ui_idxs = np.where(~np.isnan(u_interp))
                ui_fd = NearestNDInterpolator(bzyx[ui_idxs], u_interp[ui_idxs])
                ui_idxs = np.where(np.isnan(u_interp))
                u_interp[ui_idxs] = ui_fd(bzyx[ui_idxs])

            if np.any(np.isnan(u_interp)):
                raise ValueError('No boundary  u velocity data for GOFS. '
                                 'Try increasing pixel_buffer argument.')

            vi_fd = RegularGridInterpolator(
                (zi, yi, xi),
                vi,
                method='linear',
                bounds_error=False,
                fill_value=np.nan
            )

            v_interp = vi_fd(bzyx)

            if np.all(np.isnan(v_interp)):
                vi_idxs = np.where(~np.isnan(vi))
                xi = np.tile(xi, vi.shape)[vi_idxs].flatten()
                yi = np.tile(yi, vi.shape)[vi_idxs].flatten()
                zi = np.tile(zi, vi.shape)[vi_idxs].flatten()
                xyzi = np.vstack([xi, yi, zi]).T
                vi_fd = NearestNDInterpolator(xyzi, vi[ui_idxs].flatten())
                v_interp = ui_fd(np.vstack([bx, by, -bz.flatten()]).T)

            elif np.any(np.isnan(v_interp)):
                vi_idxs = np.where(~np.isnan(v_interp))
                vi_fd = NearestNDInterpolator(bzyx[vi_idxs], v_interp[vi_idxs])
                vi_idxs = np.where(np.isnan(v_interp))
                v_interp[vi_idxs] = vi_fd(bzyx[vi_idxs])

            if np.any(np.isnan(v_interp)):
                raise ValueError('No boundary data for GOFS. Try increasing pixel_buffer argument.')

            dst['time_series'][i, offset:offset+bz.shape[0], :, 0] = u_interp.reshape(bz.shape)
            dst['time_series'][i, offset:offset+bz.shape[0], :, 1] = v_interp.reshape(bz.shape)


class GOFSTemperature(GOFSComponent):

    @property
    def ncvar(self):
        return 'water_temp'

    def put_boundary_ncdata(
            self,
            hgrid,
            vgrid,
            boundary,
            dst,
            start_date,
            run_days,
            overwrite=False,
            offset=0,
            output_interval=timedelta(hours=24),
            pixel_buffer=10,
            progress_bar=True
    ):

        for i, (time, dataset) in enumerate(
            self.get_datasets(
                start_date,
                run_days,
                output_interval
            ).items()
        ):
            if start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:]),
                    '%Y-%m-%d%H:%M:%S')
                #time1 = dataset['time']
                #ds_timevector = nc4.num2date(
                #    time1,
                #    units=time1.units,
                #    only_use_cftime_datetimes=False) 
            else:
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:-1]),
                    '%Y-%m-%d%H:%M:%S.%f')
            ds_timevector = [ds_base_date + timedelta(hours=x)
                             for x in dataset['time'][:]]
            requested_date = dates.nearest_cycle(
                start_date + i*output_interval,
                period=3).replace(tzinfo=None)
            time_idx = ds_timevector.index(requested_date)
            logger.info(
                'Saving GOFS temp data for date: '
                f'{start_date+i*output_interval} '
                f'approximated as {ds_timevector[time_idx]} for '
                f'boundary id={boundary.id}'
                )
            bounds = boundary.geometry.bounds
            dx = (dataset['lon'][-1] - dataset['lon'][0]) / len(dataset['lon'])
            dy = (dataset['lat'][-1] - dataset['lat'][0]) / len(dataset['lat'])
            bounds = (
                bounds[0] - 2*dx,
                bounds[1] - 2*dy,
                bounds[2] + 2*dx,
                bounds[3] + 2*dy,
                )
            bbox = self._modified_bbox(
                dataset, Bbox.from_extents(*bounds))
            lon_idxs, lat_idxs = self._modified_bbox_indexes(
                    bbox,
                    dataset,
                    pixel_buffer
                )
            # z_ui_idxs = list(range(dataset['depth'].shape[0]))
            z_idxs = list(range(dataset['depth'].shape[0]))  # TODO: subset?
            #temp = np.full((len(z_idxs), len(lat_idxs), len(lon_idxs)), np.nan)
            if progress_bar is True:
                items_iter = tqdm.tqdm(lat_idxs)
                with tqdm_logging_wrapper.wrap_logging_for_tqdm(
                        items_iter), items_iter:
                    #for k, lat_idx in enumerate(items_iter):
                    #    temp[:, k, :] = dataset[self.ncvar][time_idx, z_idxs, lat_idx, lon_idxs]
                    temp = dataset[self.ncvar][time_idx, :, lat_idxs, lon_idxs]
                    salt = dataset['salinity'][time_idx, :, lat_idxs, lon_idxs]
            else:
                #for k, lat_idx in enumerate(lat_idxs):
                #    temp[:, k, :] = dataset[self.ncvar][time_idx, z_idxs, lat_idx, lon_idxs]
                temp = np.squeeze(dataset[self.ncvar][time_idx, :, lat_idxs, lon_idxs])
                salt = np.squeeze(dataset['salinity'][time_idx, :, lat_idxs, lon_idxs])

            #convert in-situ temperature to potential temperature
            print(f'The shape of temp is {temp.shape}')
            nz = temp.shape[0]
            ny = temp.shape[1]
            nx = temp.shape[2]
            dep = dataset['depth'][:]
            pr=np.ones(temp.shape)
            pre = pr*dep[:,None, None]
            Pr = np.zeros(temp.shape)
            ptemp = sw.ptmp(salt, temp, pre, Pr)*1.00024

            #change missing value to nan
            idxs = np.where(abs(ptemp) > 10000)
            ptemp[idxs] = float('nan')

            loni = dataset['lon'][lon_idxs]
            for idx in range(len(loni)):
                if loni[idx] > 180:
                    loni[idx] = loni[idx]-360.
            lati = dataset['lat'][lat_idxs]
            xi, yi = transform_ll_to_cpp(loni, lati)

            if vgrid.ivcor == 1:
                bz = (hgrid.values[:, None]*vgrid.sigma)[boundary.indexes, :]
                idxs = np.where(bz > 5000.0)
                bz[idxs] = 5000.0 - 1.0e-6
                print(f'zcor at 200 is {bz[200,:]}')
            else:
                raise NotImplementedError('vgrid.ivcor!=1')

            xy = hgrid.get_xy(crs='epsg:4326')
            lonb = xy[boundary.indexes, 0]
            latb = xy[boundary.indexes, 1]
            xb, yb = transform_ll_to_cpp(lonb, latb)
            #bx = np.tile(xy[boundary.indexes, 0], [bz.shape[1],1])
            #by = np.tile(xy[boundary.indexes, 1], [bz.shape[1],1])
            #bzyx = np.vstack([bz.flatten(), by, bx]).T
            bx = np.tile(xb, [bz.shape[1],1]).T
            by = np.tile(yb, [bz.shape[1],1]).T
            bzyx = np.c_[bz.reshape(np.size(bz)), by.reshape(np.size(by)), bx.reshape(np.size(bx))]
            zi = dataset['depth'][z_idxs]

            # First try with RegularGridInterpolator
            ptemp_fd = RegularGridInterpolator(
                (zi, yi, xi),
                ptemp,
                method='linear',
                bounds_error=False,
                fill_value=np.nan
            )
            ptemp_interp = ptemp_fd(bzyx)

            # the boundary and the data don't intersect
            if np.all(np.isnan(ptemp_interp)):
                ptemp_idxs = np.where(~np.isnan(ptemp))
                xyzi = np.vstack([
                    np.tile(xi, ptemp.shape)[ptemp_idxs].flatten(),
                    np.tile(yi, ptemp.shape)[ptemp_idxs].flatten(),
                    np.tile(zi, ptemp.shape)[ptemp_idxs].flatten()
                ]).T
                ptemp_fd = NearestNDInterpolator(xyzi, ptemp[ptemp_idxs].flatten())
                ptemp_interp = ptemp_fd(np.vstack([bx, by, -bz.flatten()]).T)
            # the boundary and the data partially intersect
            elif np.any(np.isnan(ptemp_interp)):
                ptemp_idxs = np.where(~np.isnan(ptemp_interp))
                ptemp_fd = NearestNDInterpolator(bzyx[ptemp_idxs], ptemp_interp[ptemp_idxs])
                ptemp_idxs = np.where(np.isnan(ptemp_interp))
                ptemp_interp[ptemp_idxs] = ptemp_fd(bzyx[ptemp_idxs])

            if np.any(np.isnan(ptemp_interp)):
                raise ValueError('No boundary  temperature data for GOFS. '
                                 'Try increasing pixel_buffer argument.')
            dst['time_series'][i, offset:offset+bz.shape[0], :, :] = ptemp_interp.reshape(bz.shape)


class GOFSSalinity(GOFSComponent):

    @property
    def ncvar(self):
        return 'salinity'

    def put_boundary_ncdata(
            self,
            hgrid,
            vgrid,
            boundary,
            dst,
            start_date,
            run_days,
            overwrite=False,
            offset=0,
            output_interval=timedelta(hours=24),
            pixel_buffer=10,
            progress_bar=True
    ):

        for i, (time, dataset) in enumerate(
            self.get_datasets(
                start_date,
                run_days,
                output_interval
            ).items()
        ):
            if start_date.strftime("%Y-%m-%d") < datetime.now().strftime("%Y-%m-%d"):
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:]),
                    '%Y-%m-%d%H:%M:%S')
                #time1 = dataset['time']
                #ds_timevector = nc4.num2date(
                #    time1,
                #    units=time1.units,
                #    only_use_cftime_datetimes=False) 
            else:
                ds_base_date = datetime.strptime(
                    ''.join(dataset['time'].units.split()[2:-1]),
                    '%Y-%m-%d%H:%M:%S.%f')
            ds_timevector = [ds_base_date + timedelta(hours=x)
                             for x in dataset['time'][:]]
            requested_date = dates.nearest_cycle(
                start_date + i*output_interval,
                period=3).replace(tzinfo=None)
            time_idx = ds_timevector.index(requested_date)
            print(f'time_idx for {requested_date} is {time_idx}')
            logger.info(
                'Saving GOFS salinity data for date: '
                f'{start_date+i*output_interval} '
                f'approximated as {ds_timevector[time_idx]} for '
                f'boundary id={boundary.id}'
                )
            bounds = boundary.geometry.bounds
            dx = (dataset['lon'][-1] - dataset['lon'][0]) / len(dataset['lon'])
            dy = (dataset['lat'][-1] - dataset['lat'][0]) / len(dataset['lat'])
            bounds = (
                bounds[0] - 2*dx,
                bounds[1] - 2*dy,
                bounds[2] + 2*dx,
                bounds[3] + 2*dy,
                )
            bbox = self._modified_bbox(
                dataset, Bbox.from_extents(*bounds))
            lon_idxs, lat_idxs = self._modified_bbox_indexes(
                    bbox,
                    dataset,
                    pixel_buffer
                )
            print(f'lon_idxs is {lon_idxs}')
            print(f'lat_idxs is {lat_idxs}')

            # z_ui_idxs = list(range(dataset['depth'].shape[0]))
            z_idxs = list(range(dataset['depth'].shape[0]))  # TODO: subset?
            #salt = np.full((len(z_idxs), len(lat_idxs), len(lon_idxs)), np.nan)
            if progress_bar is True:
                items_iter = tqdm.tqdm(lat_idxs)
                with tqdm_logging_wrapper.wrap_logging_for_tqdm(
                        items_iter), items_iter:
                    #for k, lat_idx in enumerate(items_iter):
                    #    salt[:, k, :] = dataset[self.ncvar][time_idx, z_idxs, lat_idx, lon_idxs]
                    salt = dataset[self.ncvar][time_idx, :, lat_idxs, lon_idxs]
            else:
                #for k, lat_idx in enumerate(lat_idxs):
                #    salt[:, k, :] = dataset[self.ncvar][time_idx, z_idxs, lat_idx, lon_idxs]
                salt = dataset[self.ncvar][time_idx, :, lat_idxs, lon_idxs]
            #change missing value to nan
            idxs = np.where(abs(salt) > 10000)
            salt[idxs] = float('nan')

            loni = dataset['lon'][lon_idxs]
            for idx in range(len(loni)):
                if loni[idx] > 180:
                    loni[idx] = loni[idx]-360.
            lati = dataset['lat'][lat_idxs]
            xi, yi = transform_ll_to_cpp(loni, lati)

            if vgrid.ivcor == 1:
                bz = (hgrid.values[:, None]*vgrid.sigma)[boundary.indexes, :]
                print(f'zcor is {bz[200,:]}')
                idxs = np.where(bz > 5000.0)
                bz[idxs] = 5000.0 - 1.0e-6
            else:
                raise NotImplementedError('vgrid.ivcor!=1')

            xy = hgrid.get_xy(crs='epsg:4326')
            lonb = xy[boundary.indexes, 0]
            latb = xy[boundary.indexes, 1]
            xb, yb = transform_ll_to_cpp(lonb, latb)
            bx = np.tile(xb, [bz.shape[1],1]).T
            by = np.tile(yb, [bz.shape[1],1]).T
            bzyx = np.c_[bz.reshape(np.size(bz)), by.reshape(np.size(by)), bx.reshape(np.size(bx))]
            zi = dataset['depth'][z_idxs]

            # First try with RegularGridInterpolator
            salt_fd = RegularGridInterpolator(
                (zi, yi, xi),
                salt,
                method='linear',
                bounds_error=False,
                fill_value=np.nan
            )
            salt_interp = salt_fd(bzyx)

            # the boundary and the data don't intersect
            if np.all(np.isnan(salt_interp)):
                salt_idxs = np.where(~np.isnan(salt))
                xyzi = np.vstack([
                    np.tile(xi, salt.shape)[salt_idxs].flatten(),
                    np.tile(yi, salt.shape)[salt_idxs].flatten(),
                    np.tile(zi, salt.shape)[salt_idxs].flatten()
                ]).T
                salt_fd = NearestNDInterpolator(xyzi, salt[salt_idxs].flatten())
                salt_interp = salt_fd(np.vstack([bx, by, -bz.flatten()]).T)
            # the boundary and the data partially intersect
            elif np.any(np.isnan(salt_interp)):
                salt_idxs = np.where(~np.isnan(salt_interp))
                salt_fd = NearestNDInterpolator(bzyx[salt_idxs], salt_interp[salt_idxs])
                salt_idxs = np.where(np.isnan(salt_interp))
                salt_interp[salt_idxs] = salt_fd(bzyx[salt_idxs])

            if np.any(np.isnan(salt_interp)):
                raise ValueError('No boundary  salt data for GOFS. '
                                 'Try increasing pixel_buffer argument.')
            dst['time_series'][i, offset:offset + bz.shape[0], :, :] = salt_interp.reshape(bz.shape)


class GOFS(Hycom):
    '''Public interface for GOFS model forcings.'''

    def __init__(self):
        self._elevation = GOFSElevation()
        self._velocity = GOFSVelocity()
        self._temperature = GOFSTemperature()
        self._salinity = GOFSSalinity()

    @property
    def elevation(self) -> GOFSElevation:
        return self._elevation

    @property
    def velocity(self) -> GOFSVelocity:
        return self._velocity

    @property
    def temperature(self) -> GOFSTemperature:
        return self._temperature

    @property
    def salinity(self) -> GOFSSalinity:
        return self._salinity



from time import time
import os
import argparse
from datetime import datetime, timedelta
import logging
import json

import numpy as np

from pyschism.mesh import Hgrid
from pyschism.forcing.bctides import Bctides

logging.basicConfig(
    format = "[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logging.getLogger("pyschism").setLevel(logging.DEBUG)

def list_of_strings(arg):
    return arg.split(',')

if __name__ == "__main__":

    '''
    Assume files are already located in:
    database='fes2014'
        ~/.local/share/fes2014/eastward_velocity/
        ~/.local/share/fes2014/northward_velocity/  
        ~/.local/share/fes2014/ocean_tide_extrapolated/
    database = 'tpxo'
        ~/.local/share/tpxo/
    '''
    parser = argparse.ArgumentParser(description="Create bctides.in for SCHISM with command-line arguments! e.g. python test_bctides.py hgrid.ll 2014-12-01 397 '[[5,5,4,4],[5,5,4,4],[0,1,1,2]]' major fes2014")
    
    #Add arguments
    parser.add_argument('hgrid', type=str, help='hgrid.ll (lon/lat) file')
    parser.add_argument('start_date', type=datetime.fromisoformat, help='model startdate')
    parser.add_argument('rnday', type=float, help='model rnday')
    parser.add_argument('bctypes', type=str, help="JSON format for Flags for each open boundary, '[[5,5,4,4],[5,5,4,4],[0,1,1,2]]'")
    parser.add_argument('constituents', type=list_of_strings, help="Choose tidal constituents to be included, major, minor, or list of constituents ['K1', 'O1', 'M2']")
    parser.add_argument('database', type=str, help='Tidal datbase: tpxo or fes2014')

    #Parse the command-line arguments
    args = parser.parse_args()
    hgrid_filename = args.hgrid
    start_date = args.start_date
    rnday = args.rnday
    bctypes = args.bctypes
    constituents = args.constituents
    database = args.database

    # Parse the JSON string into a Python data structure
    try:
        flags = json.loads(bctypes)
        print("Parsed bctype list:", flags)
    except json.JSONDecodeError:
        raise TypeError("Invalid JSON format for bctype list.")

    #earth tidal potential
    add_earth_tidal_potential = input("Would you like to add earth tidal potential? Y/N: ")
    if add_earth_tidal_potential.lower() == "y":
        earth_tidal_potential = True
    else:
        earth_tidal_potential = False

    #Check if constant values needed
    ethconst = []
    vthconst = []
    tthconst = []
    sthconst = []
    tobc = []
    sobc = []
    relax = []

    for ibnd, flag in enumerate(flags):
        iettype, ifltype, itetype, isatype = [i for i in flag]
        if iettype == 2:
            val = input(f"Elevation value at boundary {ibnd+1}: ")
            ethconst.append(float(val))
        else:
            ethconst.append(np.nan)

        if ifltype == 2:
            val = input(f"Discharge value (negative for inflow) at boundary {ibnd+1}: ")
            vthconst.append(float(val))
        elif ifltype == -4:
            val = input(f"Relaxation constants (between 0 and 1) for inflow at boundary {ibnd+1}: ")
            relax.append(float(val))
            val = input(f"Relaxation constants (between 0 and 1) for outflow at boundary {ibnd+1}: ")
            relax.append(float(val))
            
        else:
            vthconst.append(np.nan)

        if itetype == 2:
            val = input(f"Temperature value at boundary {ibnd+1}: ")
            tthconst.append(float(val))
            val = input(f"Nuding factor (between 0 and 1) for temperature at boundary {ibnd+1}: ")
            tobc.append(float(val))
        elif itetype == 1 or itetype == 3 or itetype == 4:
            tthconst.append(np.nan)
            val = input(f"Nuding factor (between 0 and 1) for temperature at boundary {ibnd+1}: ")
            tobc.append(float(val))
        else:
            tthconst.append(np.nan)
            tobc.append(np.nan)

        if isatype == 2:
            val = input(f"Salinity value at boundary {ibnd+1}: ")
            sthconst.append(float(val))
            val = input(f"Nuding factor (between 0 and 1) for salinity at boundary {ibnd+1}: ")
            sobc.append(float(val))
        elif isatype == 1 or isatype == 3 or isatype == 4:
            sthconst.append(np.nan)
            val = input(f"Nuding factor (between 0 and 1) for salinity at boundary {ibnd+1}: ")
            sobc.append(float(val))
        else:
            sthconst.append(np.nan)
            sobc.append(np.nan)

    outdir = './'
    hgrid = Hgrid.open(hgrid_filename, crs="epsg:4326")

    bctides=Bctides(
        hgrid = hgrid, 
        flags = flags,
        constituents = constituents, 
        database = database,
        add_earth_tidal = earth_tidal_potential,
        ethconst = ethconst,
        vthconst = vthconst,
        tthconst = tthconst,
        sthconst = sthconst,
        tobc = tobc,
        sobc = sobc,
        relax = relax,
    )

    bctides.write(
        outdir, 
        start_date=start_date, 
        rnday=rnday, 
        overwrite=True,
    )
